/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include "Camera.h"
#include <glm/gtx/rotate_vector.hpp>
#include <time.h>
#include <cstring>

#include <math.h>
#define TRACKBALLSIZE  (0.8)
#define AERIALVIEWLIMIT 89.9 * 14
static float tbProjectToSphere(float, float, float);
static void normalizeQuat(float [4]);
#pragma warning(disable : 4996)
const float PI = float(atan(1.0) * 4.0);

CPhysicalCamera::CPhysicalCamera()
    : fSetTime(0), fTotalMoveTime(0), vMove(1.0f)
    , bRoate_camera_flag(false), bMove_camera_flag(false)
    , fRotationAngleCube(0), fRotationAnglePyramid(0)
    , fCubeRotationSpeed(0), fPyramidRotationSpeed(0)
    , PIover180(0), iForw(0), iBack(0), iLeft(0)
    , iRight(0), flength(0), bRotateCameraModel(false)
    , AerialViewY(0), bAerialView(false)
    , fRotateCameraXSpeed(0.0), fRotateCameraYSpeed(0.0)
    , bAontrolCamera(false), bMoveCmeraModel(false)
    , mRotateY(1.0f), mRotateX(1.0f)
    , vSpeed(1.0f), mLook(1.0f), mProjection(1.0f)
    , vModelCenterPos(1.0f), viewMatrix(1.0f)
{
    vEye = glm::vec3(0.0f, 0.0f, 0.0f);
    vView = glm::vec3(0.0f, 0.0, -1.0f);
    vUp = glm::vec3(0.0f, 1.0f, 0.0f);
    fSpeed = 25.0f;
    fSensitivity = 0.1f;
    memset(lastquat,0,sizeof(float)*4);
    memset(curquatY,0,sizeof(float)*4);
    memset(curquatX,0,sizeof(float)*4);
}

CPhysicalCamera::CPhysicalCamera(glm::vec3 a_vEye,
                                 glm::vec3 a_vView,
                                 glm::vec3 a_vUp,
                                 double a_fSpeed,
                                 double a_fSensitivity)
    : fSetTime(0), fTotalMoveTime(0), vMove(1.0f)
    , bRoate_camera_flag(false), bMove_camera_flag(false)
    , fRotationAngleCube(0), fRotationAnglePyramid(0)
    , fCubeRotationSpeed(0), fPyramidRotationSpeed(0)
    , PIover180(0), iForw(0), iBack(0), iLeft(0)
    , iRight(0), flength(0), bRotateCameraModel(false)
    , AerialViewY(0), bAerialView(false)
    , fRotateCameraXSpeed(0.0), fRotateCameraYSpeed(0.0)
    , bAontrolCamera(false), bMoveCmeraModel(false)
    , mRotateY(1.0f), mRotateX(1.0f)
    , vSpeed(1.0f), mLook(1.0f), mProjection(1.0f)
    , vModelCenterPos(1.0f), viewMatrix(1.0f)
{
    vEye = a_vEye; vView = a_vView; vUp = a_vUp;
    fSpeed = a_fSpeed;
    fSensitivity = a_fSensitivity;
    memset(lastquat,0,sizeof(float)*4);
    memset(curquatY,0,sizeof(float)*4);
    memset(curquatX,0,sizeof(float)*4);
}

void CPhysicalCamera::buildRotMatrix(glm::mat4& m, float q[4])
{
    m[0][0] = (float)(1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]));
    m[0][1] = (float)(2.0 * (q[0] * q[1] - q[2] * q[3]));
    m[0][2] = (float)(2.0 * (q[2] * q[0] + q[1] * q[3]));
    m[0][3] = 0.0;

    m[1][0] = (float)(2.0 * (q[0] * q[1] + q[2] * q[3]));
    m[1][1] = (float)(1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]));
    m[1][2] = (float)(2.0 * (q[1] * q[2] - q[0] * q[3]));
    m[1][3] = 0.0;

    m[2][0] = (float)(2.0 * (q[2] * q[0] - q[1] * q[3]));
    m[2][1] = (float)(2.0 * (q[1] * q[2] + q[0] * q[3]));
    m[2][2] = (float)(1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]));
    m[2][3] = 0.0;

    m[3][0] = 0.0;
    m[3][1] = 0.0;
    m[3][2] = 0.0;
    m[3][3] = 1.0;
}

float CPhysicalCamera::GetAngleX()
{
    glm::vec3 vDir = vView - vEye;
    vDir = glm::normalize(vDir);
    glm::vec3 vDir2 = vDir; vDir2.y = 0.0f;
    vDir2 = glm::normalize(vDir2);
    float fAngle = acos(glm::dot(vDir2, vDir)) * (180.0f / PI);
    if (vDir.y < 0)
    {
        fAngle *= -1.0f;
    }
    return fAngle;
}

void CPhysicalCamera::rotateCamera(double horizontal, double vertical,
                                   double planar, double time)
{
    if (time != 0)
    {
        bRoate_camera_flag = true;
        fSetTime = time;
        glm::vec3 endView(horizontal, vertical, planar);
        glm::vec3 vRotate = endView - vView;
        vRotate /= time;
        vSpeed = vRotate;
    }
    else
    {
        float deltaX = (float)(horizontal * fSensitivity);
        float deltaY = (float)(vertical * fSensitivity);
        glm::vec3 vCameraRotate = vView;
        if (deltaX != 0.0f)
        {
            vCameraRotate -= vEye;
            vCameraRotate = glm::rotate(vCameraRotate, deltaX,
                                        glm::vec3(0.0f, 1.0f, 0.0f));
            vCameraRotate += vEye;
        }
        if (deltaY != 0.0f)
        {
            glm::vec3 vAxis = glm::cross(vCameraRotate - vEye, vUp);
            vAxis = glm::normalize(vAxis);
            float fAngle = deltaY;
            float fNewAngle = fAngle + GetAngleX();
            if (fNewAngle > -89.80f && fNewAngle < 89.80f)
            {
                vCameraRotate -= vEye;
                vCameraRotate = glm::rotate(vCameraRotate, deltaY, vAxis);
                vCameraRotate += vEye;
            }
        }
        vView = vCameraRotate;
    }
}

void CPhysicalCamera::moveCamera(double x, double y, double z,
                                 double time)
{

    if (time != 0)
    {
        bMove_camera_flag = true;
        fSetTime = time;
        glm::vec3 endView(x, y, z);
        glm::vec3 vRotate = endView - vView;
        vRotate /= time ;
        vSpeed = vRotate;
    }
    else
    {
        if (abs(x) > 0.0001 || abs(y) > 0.0001 || abs(z) > 0.0001)
        {
             bMoveCmeraModel = true;
        }
        vMove = glm::vec3(x, y, z);
        fSetTime = 0.0;
    }
}

void CPhysicalCamera::SetRotationAngle(double fRotateCube ,
                                       double fRotatePyramid ,
                                       double fCubeRotateSpeed,
                                       double fPyramidRotateSpeed)
{
    PIover180 = 3.1415f / 180.0f;
    fRotationAngleCube = fRotateCube;
    fRotationAnglePyramid = fRotatePyramid;
    fCubeRotationSpeed = fCubeRotateSpeed;
    fPyramidRotationSpeed = fPyramidRotateSpeed;
}
 void CPhysicalCamera::setLook(glm::mat4 m)
{
    viewMatrix = m;
}

glm::mat4& CPhysicalCamera::getLook()
{
    return viewMatrix;
}

void CPhysicalCamera::setPerspective(glm::mat4 Perspective)
{
    mProjection = Perspective;
}

const glm::mat4 CPhysicalCamera::getPerspective()
{
    return mProjection;
}

glm::mat4 CPhysicalCamera::getModelViewMatrix(glm::mat4 lookat,
                                              glm::mat4& matrix,
                                              float fMoveTime,
                                              float fPreTime)
{
    glm::mat4 modelViewMatrix;
    if (fPreTime <= 0)
    {
        modelViewMatrix = lookat * matrix;
    }
    else
    {
        fTotalMoveTime += fMoveTime;
        if (fTotalMoveTime <= fSetTime)
        {
            if (bRoate_camera_flag == true)
            {
                vView += vSpeed * fMoveTime;
            }
            if (bMove_camera_flag == true)
            {
                vEye += vSpeed * fMoveTime;
                vView += vSpeed * fMoveTime;
            }
            modelViewMatrix = lookat * matrix;
        }
        else
        {
            modelViewMatrix = lookat * matrix;
        }
    }
    return modelViewMatrix;
}

void CPhysicalCamera::RotateObjectMouse(double horizontal, double vertical,
                                        double planar)
{
    if (!bAerialView)
        return;
    if (abs(horizontal) > planar || abs(vertical) > 0.0001)
    {
        bRotateCameraModel = true;
        float deltaX = (float)(horizontal * fSensitivity * 0.01f);
        float deltaY = (float)(-vertical * fSensitivity * 0.01f);
        AerialViewY += -vertical;
        if (AerialViewY >= 0.0 && AerialViewY <= AERIALVIEWLIMIT)
        {
            trackball(lastquat, 0.0, deltaY, 0.0, 0.0);
            add_quats(lastquat, curquatY, curquatY);
        }
        else
        {
            AerialViewY -= -vertical;
        }
        trackball(lastquat, 0.0f, 0.0f, 0.0f, 0.0f);
        trackball(lastquat, deltaX, 0.0, 0.0, 0.0);
        add_quats(lastquat, curquatX, curquatX);
    }
}

void CPhysicalCamera::removeRotateObject()
{
    trackball(lastquat, 0.0f, 0.0f, 0.0f, 0.0f);
    fCubeRotationSpeed = 0.0f;
    fPyramidRotationSpeed = 0.0f;
    fCubeRotationSpeed = 0.0f;
    fPyramidRotationSpeed = 0.0f;
}

void CPhysicalCamera::RotateCamera(double horizontal, double vertical,
                                   double planar)
{
    if (abs(horizontal) > planar || abs(vertical) > 0.0001)
    {
        bRotateCameraModel = true;
    }
    fRotateCameraXSpeed = 0.0f;
    fRotateCameraYSpeed = planar;
    double deltaX = horizontal * fSensitivity;
    double deltaY = vertical * fSensitivity;
    if(deltaX != 0.0f)
    {
        fRotateCameraXSpeed =- deltaX * 3;
    }
    if(deltaY != 0.0f)
    {
        fRotateCameraYSpeed =- deltaY * 3;
    }
}
