/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include "FPSCounter.h"
#if ENABLE_FPS
bool FPSCounter::loadFPSShader(ShaderProgram* pShaderProgram)
{
    uiFPSProgram = glCreateProgram();
    pShaderProgram->loadShader(uiFPSProgram, VERTSHADER, strlen(VERTSHADER),
                               GL_VERTEX_SHADER);
    pShaderProgram->loadShader(uiFPSProgram, FRAGSHADER, strlen(FRAGSHADER),
                               GL_FRAGMENT_SHADER);
    pFont->setShaderProgram(uiFPSProgram);
    return true;
}

void FPSCounter::init(ShaderProgram* pShaderProgram)
{
    loadFPSShader(pShaderProgram);
#ifdef _WIN32
    QueryPerformanceFrequency(&f);
    dqFreq = (double)f.QuadPart;
#endif

#if WRITEFPS2FILE
    timeCounter = 0;
#endif
}

void FPSCounter::timeStamp()
{
    mLastTime = mCurrentTime;
#ifdef _WIN32
    _LARGE_INTEGER tmp;
    QueryPerformanceCounter(&tmp);
    mCurrentTime = (double)tmp.QuadPart;
#else
    mCurrentTime = clock();
#endif
}

double FPSCounter::getExecutionTime()
{
    double renderTime = 0;
    if (mLastTime != 0)
    {
#ifdef _WIN32
        renderTime = (double)(mCurrentTime - mLastTime) / dqFreq;
#else
        renderTime = (double)(mCurrentTime - mLastTime) / CLOCKS_PER_SEC;
#endif
    }
    return renderTime * 1000;
}
#if QA_TEST
static int countTime=0;
#endif

void FPSCounter::printFPS(glTFViewport* pViewport)
{
    glUseProgram(uiFPSProgram);
    glDisable(GL_DEPTH_TEST);
    GLuint iLoc = glGetUniformLocation(uiFPSProgram, "projMatrix");
    const glm::mat4 aMat = glm::ortho(0.0f, float(pViewport->width), 0.0f,
                                      float(pViewport->height));
    glUniformMatrix4fv(iLoc, 1, false, glm::value_ptr(aMat));
    glm::vec4 lightVector = glm::vec4(0.0f, 1.0f, 1.0f, 1.0f);
    iLoc = glGetUniformLocation(uiFPSProgram, "vColor");
    glUniform4fv(iLoc, 1, (GLfloat*)&lightVector);
    
    if (mDurationTime >= 1000)
    {
        mFPS = mFrames;
        mDurationTime = 0;
        mFrames = 0;
        mTotalTime = getExecutionTime();

#if WRITEFPS2FILE
        timeCounter++;
        if (timeCounter >= TIMETHRESHOLD)
        {
            FPSRecoder.push_back(mFPS);
            timeRecorder.push_back(mTotalTime);
            timeCounter = 0;
        }
#endif

#if QA_TEST
        countTime++;
        if(countTime >= AUTO_CLOSE_TIME)
        {
            FPSCounter::~FPSCounter();
            exit(0);
        }
#endif
    }
    else
    {
        mDurationTime += getExecutionTime();
        ++mFrames;
    }

    pFont->printFormattedString(pViewport->width - 100, 30, 15, "%d", mFPS);

    glEnable(GL_DEPTH_TEST);
}
#endif //ENABLE_FPS
